#include <dos.h>                        // FP_SEG() FP_OFF() getvect() setvect() enable() disable()
#include <conio.h>                      // getch() kbhit()
#include <stdio.h>                      // printf()
#include <stdlib.h>                     // exit()
#include <stddef.h>                     // offsetof()
#include <string.h>                     // strstr()
#include <mem.h>                        // memset()
#include <errno.h>                      // extern errno

#undef outportb
#undef outportw
#undef inportb
#undef inportw

#include "..\dpmi32\phys.h"
#include "..\dpmi32\intr.h"
#include "..\pci\pci.h"
#include "..\intel\timestmp.h"          // class PentiumClock
#include "..\intel\i8259a.h"            // class InterruptController
#include "..\intel\i8254-2.h"           // class HardwareTimer
#include "..\aal5\atm.h"                // struct AtmHeader
#include "..\fore\pca200e.h"
#include "..\intel\byteswap.h"          // ntohl()



// conditional compilation switches
//
#define ATM_INIT_RESETS_SAR             TRUE
#define UART_TALKS                      TRUE
#define UART_LISTENS                    TRUE
#define SWAP_FIRMWARE                   FALSE

#define ALLOW_SONET_INTERRUPTS          FALSE
#define ALLOW_ATM_INTERRUPTS            FALSE
#define ALLOW_ATM_RX_FREE_EMPTY_IRQ     TRUE
#define ALLOW_TIMER_INTERRUPTS          FALSE       // FALSE to debug handler code
#define AFFECT_TIMER_INTERRUPTS         TRUE
#define ESSENTIAL_STATUS_QUERIES        TRUE
#define ATM_EXIT_CLEARS_RX_DMA_STATES   FALSE
#define ATM_EXIT_ITERATES_TIMER_HANDLER FALSE       // TRUE to test w/o interrupts
#define ATM_EXIT_CALLS_ATM_SHIFT        TRUE        // FALSE to see interrupt handler timings when done
#define ATM_EXIT_RELEASES_BLOCKS        TRUE
#define ATM_EXIT_DUMPS_ATM_HEADERS      FALSE
#define ATM_INIT_TRACE                  FALSE
#define ATM_INIT_TRACE2                 FALSE
#define ATM_INIT_EXTRA_STATUS_CHECKS    FALSE
#define ATM_INIT_SETS_PCI_TIMING        FALSE
#define ATM_INIT_SETS_CACHING           FALSE
#define ATM_POPULATE_TALKS              FALSE
#define ATM_POPULATE_EXITS_EARLY        FALSE
#define ATM_SHIFT_TALKS                 TRUE
#define ATM_SHIFT_CLEANUP_DEBUG         FALSE
#define ATM_HANDLER_BYPASSED            FALSE
#define SONET_INIT_RESETS_FRAMER        TRUE
#define SAVE_INTERRUPT_THIS_PTR         FALSE
#define SWAP_ATM_HEADER                 TRUE
#define CAPTURE_RAW_CELLS               FALSE
#define FORCE_EOP_WAIT                  TRUE
#define FORCE_PACKET_RESYNC             FALSE
#define PCI_INIT_STOPS_EARLY            FALSE
#define PCI_INIT_SETS_PCI_TIMING        TRUE
#define PCI_INIT_SETS_CACHING           FALSE
#define CHECK_MEMORY_CONSISTENCY        TRUE        // no time wasted except on startup and shutdown
#define CHECK_MEMORY_WHILE_RUNNING      FALSE       // extra checks while >=1 card still running, affects RX free ring unavail count
#define CHECK_MEMORY_WHILE_LOOPING      FALSE       // extra checks inside main loop, big performance hit
#define COMPL_AND_FREE_IN_LOCKSTEP      TRUE
#define SW_RING_FITS_EVENLY_IN_HW       TRUE
#define DISABLE_TIMER_REENTRY           FALSE       // thought it would allow debugger to single-step interrupt handler, but it doesn't
#define MULTIPLE_STAMPS_PER_BLOCK       TRUE
#define LOCK_PAGES                      FALSE
#define DO_FOREGROUND_STATS             TRUE
#define PERIODIC_STATS_QUERY            FALSE
#define ALLOW_STATS_CALLS               FALSE       // main call to flows_per_block()
#define ALLOW_STATS_CALLS_2             TRUE        // extra call to flows_per_block() when empty queue
#define ALLOW_STATS_CALLS_3             TRUE        // call to flows_query()
#define BUSYWAIT_STATS_CALLS            TRUE
#define BUSYWAIT_STATS_CALLS_2          FALSE       // not implemented
#define BUSYWAIT_STATS_CALLS_3          FALSE
#define TIME_STATS_CALLS                TRUE        // main call to flows_per_block()
#define TIME_STATS_CALLS_2              FALSE       // extra call to flows_per_block() when empty queue
#define TIME_STATS_CALLS_3              FALSE       // call to flows_query()
#define ONE_CARD_PER_TIMER_IRQ          TRUE
#define SIMULATE_FIBER                  FALSE       // when TRUE, don't hook up any physical fiber
#define PCI_EXIT_CALLS_BUFFER_DUMP      FALSE       // FALSE for quick turnaround when timing



// manifest constants
//

#define MAX_ATM_NICS                2           // how many cards we expect in host computer

//#define TIMER_FREQUENCY   65536       // should be every 55.07 milliseconds
//#define TIMER_FREQUENCY   12100       // is every 10.063 milliseconds
//#define TIMER_FREQUENCY   11900       // should be every 10.0 milliseconds
//#define TIMER_FREQUENCY   3840        // is every 3.21 milliseconds
//#define TIMER_FREQUENCY   3500        // is every 2.79 milliseconds
//#define TIMER_FREQUENCY   1279        // is every .8542 milliseconds
//#define TIMER_FREQUENCY   1280        // is every 1.066 milliseconds
//#define TIMER_FREQUENCY   1190        // should be every 1.00 milliseconds, is .858 mS
//#define TIMER_FREQUENCY   900         // is every .642 milliseconds (also)
//#define TIMER_FREQUENCY   800         // is every .642 milliseconds
//#define TIMER_FREQUENCY   700         // is every .429 milliseconds     ***
#define TIMER_FREQUENCY   500         // is every .214 milliseconds
//#define TIMER_FREQUENCY   300         // is every .208 microseconds
//#define TIMER_FREQUENCY   250         // is every .208 microseconds

#define TIMER_CALIBRATE     3           // record start timestamp here

#define ATM_CALIBRATE       500         // record interrupt timestamp here

#define SONET_RESET_DELAY   60          // 1/2 microsecond worth of 120 MHz Pentium clocks
#define ATM_RESET_DELAY     64.0        // count of PCI clocks
#define ROM_START_DELAY     15.0        // count of (floating point) seconds
#define FIRM_START_DELAY    5.0         // count of (floating point) seconds
#define BEAT_CHANGE_DELAY   1.0         // count of seconds
#define UART_DELAY          5.0         // count of seconds

#define QUERY_MILLISECS     10000       // how often should we simulate remote queries? (milliseconds)
#define FLOW_BLOCK_TIME     400         // how long should dummy flows_per_block() take? (microseconds)
//#define FLOW_QUERY_TIME     56          // how long should dummy flows_query() take? (milliseconds)
#define FLOW_QUERY_TIME     0           // how long should dummy flows_query() take? (milliseconds)

#define CELL_RATE           (353207.5)

#define DESIRED_MIN_GRANT   2

#define DESIRED_CACHE_CONTROL    0x01   // (default w/o cache card) L1 cache enabled, L2 cache not forcing miss, NA# signal enabled, SRAM is 0K pipe burst
//#define DESIRED_CACHE_CONTROL    0xB1   // (default with cache card installed) L1 cache enabled, L2 cache not forcing miss, NA# signal enabled, SRAM is 512K dual bank pipe burst
//#define DESIRED_CACHE_CONTROL    0xB0   // L1 cache disabled, L2 cache disabled, NA# signal enabled, SRAM is 512K dual bank pipelined burst
//#define DESIRED_CACHE_CONTROL    0xB3   // L1 cache enabled, L2 cache is forcing miss, NA# signal enabled, SRAM is 512K dual bank pipelined burst

//#define DESIRED_HOST_CONTROL   0xE0   // CPU inactivity timer = 8 PCI clocks
#define DESIRED_HOST_CONTROL     0x80   // (default) CPU inactivity timer = 5 PCI clocks
//#define DESIRED_HOST_CONTROL   0x40   // CPU inactivity timer = 3 PCI clocks
//#define DESIRED_HOST_CONTROL   0x20   // CPU inactivity timer = 2 PCI clocks
//#define DESIRED_HOST_CONTROL   0x00   // CPU inactivity timer = 1 PCI clocks

#define DESIRED_HOST_LATENCY        0   // works well when card latency is also 0
//#define DESIRED_HOST_LATENCY      8   // works just as badly as 64, regardless of card latency
//#define DESIRED_HOST_LATENCY     64   // AMI Triton BIOS default

#define DESIRED_CARD_LATENCY        0   // only setting that allows 2 cards to keep up with IP-header-only capture
//#define DESIRED_CARD_LATENCY      8   // raw capture uses 10 RX FIFO slots with buf=4, 48-byte cells
//#define DESIRED_CARD_LATENCY     16   // raw capture uses 7 RX FIFO slots with buf=4, 48-byte cells
//#define DESIRED_CARD_LATENCY     32   // raw capture uses 6 RX FIFO slots with buf=4, 48-byte cells
//#define DESIRED_CARD_LATENCY     64   // (default) raw capture uses 5 RX FIFO slots with buf=4, 48-byte cells
//#define DESIRED_CARD_LATENCY    128   // raw capture uses 5 RX FIFO slots with buf=4, 48-byte cells
//#define DESIRED_CARD_LATENCY    248   // raw capture uses 5 RX FIFO slots with buf=4, 48-byte cells

// how much addressible memory does card require?
//
#define CARD_MEM_FOOTPRINT              sizeof(pca200e_memmap)



// error return codes
//
#define ERROR_CARD_BASE                 2000
#define ERROR_CARD_SHIFT_TOO_SLOW       (ERROR_CARD_BASE + 0)
#define ERROR_CARD_CANNOT_OPEN_FILE     (ERROR_CARD_BASE + 1)
#define ERROR_CARD_CANNOT_READ_HDR      (ERROR_CARD_BASE + 2)
#define ERROR_CARD_CANNOT_READ_BODY     (ERROR_CARD_BASE + 2)
#define ERROR_CARD_CANNOT_SEEK_END      (ERROR_CARD_BASE + 3)
#define ERROR_CARD_CANNOT_SEEK_BEG      (ERROR_CARD_BASE + 4)
#define ERROR_CARD_CANNOT_TELL_SIZE     (ERROR_CARD_BASE + 5)
#define ERROR_CARD_BSTAT_STUCK          (ERROR_CARD_BASE + 6)
#define ERROR_CARD_CP_NOT_RUNNING       (ERROR_CARD_BASE + 7)
#define ERROR_CARD_SELF_TEST_FAILED     (ERROR_CARD_BASE + 8)
#define ERROR_CARD_SELF_TEST_UNKNOWN    (ERROR_CARD_BASE + 9)
#define ERROR_CARD_HEARTBEAT_STUCK      (ERROR_CARD_BASE + 10)



unsigned timer_interrupt_card_index = 0;
unsigned long timer_interrupt_count = 0;
const unsigned timer_irq_level = 0;
unsigned char timer_interrupt;
vfp timer_thunk;
Bit48 old_timer_handler;
PentiumClock timer_start_time;
PentiumClock timer_end_time;
InterruptController *irq = 0;
HardwareTimer *timer = 0;
class AtmCard *cards[ MAX_ATM_NICS ];
Bit8 old_pci_latency_timer;
static union { PciControl old_pci_control; Bit8 old_pci_control_alias; };
static union { PciCacheControl old_pci_cache_control; Bit8 old_pci_cache_control_alias; };
#if SAVE_INTERRUPT_THIS_PTR
    class AtmCard *atm_this_pointer_during_interrupt=(AtmCard *)0x12345678;         // debug only!
#endif



// all the data that is per-card
//
class AtmCard
{

private:
    enum AtmShiftTiming
    {
        during_initialization,
        during_interrupt,
        during_cleanup
    };

    Bit32 phys_handle;
    Bit32 phys_address;
    Bit8 old_pci_latency_timer;
    pca200e_memmap *memmap;
    unsigned irq_level;
    unsigned char interrupt_number;
    vfp thunk;
    Bit48 old_handler;
    PentiumClock atm_enabled;
    PentiumClock before;
    PentiumClock elapsed;
    PentiumClock elapsed_tot;
    PentiumClock elapsed_min;
    PentiumClock elapsed_max;
    int atm_interrupt_error;
    int timer_interrupts_allowed;
    unsigned atm_out_fifo_full;
    unsigned atm_in_fifo_almost_full;
    unsigned long atm_interrupt_count;
    unsigned long timer_interrupt_count;
    union { HostControl_read atm_status; Bit32 atm_status_alias; };

    // nonstatic card interrupt handler that does the real work
    //
    int handler( void );

    // separate handlers for up to 2 cards, so we can identify
    // which card caused the interrupt
    //
    static int handler0(
        Bit32 ebp,
        Bit32 edi,
        Bit32 esi,
        Bit32 ds,
        Bit32 es,
        Bit32 edx,
        Bit32 ecx,
        Bit32 ebx,
        Bit32 eax,
        Bit32 eip,
        Bit32 cs,
        Bit32 flags
    )
    {
#if ATM_HANDLER_BYPASSED
        flat_put_long( 0xFFFBE008, 0 );
        InterruptController::acknowledge( 3 );
        return 0;
#else
        return cards[ 0 ]->handler();
#endif
    }

    static int handler1(
        Bit32 ebp,
        Bit32 edi,
        Bit32 esi,
        Bit32 ds,
        Bit32 es,
        Bit32 edx,
        Bit32 ecx,
        Bit32 ebx,
        Bit32 eax,
        Bit32 eip,
        Bit32 cs,
        Bit32 flags
    )
    {
        return cards[ 1 ]->handler();
    }

    // re-route the card interrupt to our code
    //
    // note that we must set global variable cards[cardnum]=this
    // rather than relying upon our caller to do it for us,
    // because if an interrupt occurs after we enable them,
    // but before he has assigned to global, handler0() will
    // read garbage (or zero) for the object pointer, and we
    // will corrupt host memory and probably lock up!
    //
    // it is safe for caller to assign to same global,
    // since he is not really changing the value there
    //
    void interrupt_init( unsigned cardnum )
    {
        interrupt_number = irq_to_intr( irq_level );
        old_handler = getvect( interrupt_number );
        disable();

#if ALLOW_ATM_INTERRUPTS || ALLOW_SONET_INTERRUPTS
        if( cardnum == 0 )
        {
            thunk = make_thunk( handler0, &old_handler );
        }
        else if( cardnum == 1 )
        {
            thunk = make_thunk( handler1, &old_handler );
        }
        else
        {
            printf( "fatal internal error!  cannot set vector for more than 2 ATM cards\n" );
            exit( 123 );
        }
#if THROW_IS_LEGAL
        if( thunk == NULL )
            throw "error!  cannot allocate card interrupt thunk!\n";
#endif

        Bit48 pointer;
        pointer.offset = (Bit32) thunk;
        pointer.selector = _CS;
        setvect( interrupt_number, pointer );

        cards[ cardnum ] = this;        // very important!  see above
        irq->unmask( irq_level );
#else
        irq->mask( irq_level );
#endif
        enable();
    }

    void interrupt_exit( void )
    {
        disable();
#if ALLOW_ATM_INTERRUPTS || ALLOW_SONET_INTERRUPTS
        setvect( interrupt_number, old_handler );
        free( thunk );
#endif
        enable();
    }

    void atm_uart_io( const char *output, char *input );

    int atm_download( void );

    int atm_check_heartbeat( int want_increasing );

    int atm_init( void );

    void atm_exit( void );

    void atm_handler( void );

public:
    // this field is public so main() can simulate a packet from us
    //
    // maybe we should use an accessor function instead?
    //
    Bit16 pci_handle;

    // this field is public so it can be passed to the
    // foreground stats code when a remote query occurs,
    // so the query will reflect all data seen so far,
    // in case it will be 10 days until the block fills
    //
    // maybe we should use an accessor function instead?
    //

    // constructor
    //
    AtmCard(
        Bit16 pci_handle0,
        Bit32 new_phys_address,
        unsigned new_irq_level,
        unsigned cardnum
    )
    :
        pci_handle( pci_handle0 ),
        phys_address( new_phys_address ),
        irq_level( new_irq_level ),
        timer_interrupts_allowed( FALSE )
    {
        int result;

        // get addressibility for card hardware
        //
        memmap = (pca200e_memmap *) map_physical( phys_address, CARD_MEM_FOOTPRINT, &phys_handle );

        // debug
        printf( "card 0x%4.4X is at logical address 0x%8.8lX\n", pci_handle, memmap );

#if ESSENTIAL_TIME_STAMPS
        elapsed_min.high = -1;
        elapsed_min.low  = -1;
        elapsed_max.high = 0;
        elapsed_max.low  = 0;
        elapsed_tot.high = 0;
        elapsed_tot.low  = 0;
        timer_interrupt_count = 0;
#endif

#if 0
        interrupt_init( cardnum );
#endif

        result = atm_init();
        printf( "card 0x%4.4X ATM initialized with result %d\n", pci_handle, result );
    }

    // destructor
    //
    ~AtmCard()
    {
        // clean up ATM SAR chip
        //
        atm_exit();

        // clean up interrupt grabbing
        //
        interrupt_exit();

        // give up addressibility for card
        //
        unmap_physical( CARD_MEM_FOOTPRINT, phys_handle );
    }
};



// deals with interrupts from any of 2 sources
// (SONET framer chip and/or ATM SAR chip) and 2 cards
// (monitoring send and receive fibers for a single link)
//
// since we poll the shared memory, we only use
// this interrupt to report severe errors
//
int AtmCard::handler( void )
{
#if ESSENTIAL_TIME_STAMPS
#endif

    // if ATM SAR wants attention...
    //
    atm_handler();

    // tell the interrupt controllers we are done
    //
    irq->acknowledge( irq_level );

#if ESSENTIAL_TIME_STAMPS
#endif

    // tell the thunking code not to chain to the old handler
    //
    return 0;
}



// deals with interrupts from the Texas Instruments
// TNETA1561 ATM SAR
//
// we actually plan to do the kickoff of DMA from
// timer_handler() so we can tune its granularity
//
// so this routine should only be handling errors
// (if we decide to enable those interrupts) like
// buffer overrun, the same as sonet_handler() does
//
void AtmCard::atm_handler( void )
{
#if SAVE_INTERRUPT_THIS_PTR
    atm_this_pointer_during_interrupt = this;
#endif

    // acknowledge the interrupt
    //
    union { HostControl_write hcr_w; Bit32 hcr_w_alias; };
    hcr_w_alias = 0;
    hcr_w.clear_host_int = TRUE;
    memmap->dword_00h = hcr_w_alias;

    // after I modify firmware,
    // maybe place the block of many buffers on the host filled list
    // ???

    // count how many times this occurred
    //
    atm_interrupt_count++;
}



void AtmCard::atm_uart_io( const char *output, char *input )
{
    char c;
    Soft_UART *uart = &memmap->shared.mon960.uart;
    PentiumClock now, initial, alarm;
    double timing;

    initial.set();
    now = initial;
    alarm.put_seconds( UART_DELAY );
    alarm += initial;

    // as long as we have anything to input or output...
    //
    while( now < alarm )
    {
        now.set();

        // if there is input ready for us...
        //
        if( CAN_GETC_i960( uart ) )
        {
            // read it into a local variable
            //
            GETC_i960( uart, c );

#if UART_TALKS
            // show on screen
            //
            putchar( c );
#endif

            // if the caller wants input, give it to them
            //
            if( input )
                *input++ = c;

            // reset timer
            //
            initial = now;
            alarm.put_seconds( UART_DELAY );
            alarm += initial;
        }

        // if there is room in output buffer...
        //
        if( CAN_PUTC_i960(uart,*output) )
        {
            // if the caller wants to output
            // and we haven't done all of that yet...
            //
            if( output && *output )
            {
                c = *output++;
            }
#if UART_LISTENS
            else if( kbhit() )
            {
                c = getch();
                if( c == 27 )           // escape
                    break;
            }
#endif
            else
                c = 0;

            if( c )
            {
                // spit out one character
                //
                PUTC_i960( uart, c );

                // reset timer
                //
                initial = now;
                alarm.put_seconds( UART_DELAY );
                alarm += initial;
            }
        }
    }

    // terminate the input string
    //
    if( input )
        *input++ = '\0';
}



// download i960 code image to card
//
int AtmCard::atm_download( void )
{
    FILE *handle;
    FirmwareHeader header;
    int stdio_result;
    Bit8 *start;
    Bit32 *swapper;
    long size, count;

    // disable cr-lf to lf translation!
    //
    handle = fopen( FIRMWARE_FILE_NAME, "rb" );
    if( handle == NULL )
    {
        printf( "card 0x%4.4X error %d opening file '%s'\n", pci_handle,
            errno, FIRMWARE_FILE_NAME );
        return ERROR_CARD_CANNOT_OPEN_FILE;
    }

    // get the file size
    //
    stdio_result = fseek( handle, 0, SEEK_END );
    if( stdio_result != 0 )
    {
        printf( "card 0x%4.4X error %d seeking to end of file '%s'\n",
            pci_handle, errno, FIRMWARE_FILE_NAME );

        fclose( handle );
        return ERROR_CARD_CANNOT_SEEK_END;
    }

    size = ftell( handle );
    if( size == -1L )
    {
        printf( "card 0x%4.4X error %d getting file '%s' position\n",
            pci_handle, errno, FIRMWARE_FILE_NAME );

        fclose( handle );
        return ERROR_CARD_CANNOT_TELL_SIZE;
    }

    // do we need to skip to a certain position in the file
    // to read the header???  it seems to be at 0x44 not 0...
    //
    fseek( handle, 0, SEEK_SET );
    if( stdio_result != 0 )
    {
        printf( "card 0x%4.4X error %d seeking to beginning of file '%s'\n",
            pci_handle, errno, FIRMWARE_FILE_NAME );

        fclose( handle );
        return ERROR_CARD_CANNOT_SEEK_BEG;
    }

    // read in the file header
    //
    stdio_result = fread( &header, sizeof(FirmwareHeader), 1, handle );
    if( stdio_result != 1 )
    {
        printf( "card 0x%4.4X error %d reading file '%s' %d byte header; got %d items\n",
            pci_handle, errno, FIRMWARE_FILE_NAME, sizeof(FirmwareHeader), stdio_result );

        fclose( handle );
        return ERROR_CARD_CANNOT_READ_HDR;
    }

    // we don't need to byte swap the values
    // because host (Pentium) and card (i960)
    // are both little-endian
    //
    // but if we ever do, make sure not to swap
    // again later along with rest of firmware image
    //

    // shortcut for the load address
    //
    start = memmap->ram8 + header.start_offset;
    printf( "card 0x%4.4X firmware image will load at 0x%8.8lX and start at 0x%8.8lX\n", pci_handle, start, memmap->ram8+header.entry_address );

    // copy the header to card memory
    //
    memcpy( start, &header, sizeof(FirmwareHeader) );

    // read the rest of the file directly into card memory
    //
    stdio_result = fread( start+sizeof(FirmwareHeader), size-sizeof(FirmwareHeader), 1, handle );
    if( stdio_result != 1 )
    {
        printf( "card 0x%4.4X error %d reading file '%s' %d byte body; got %d items\n",
            pci_handle, errno, FIRMWARE_FILE_NAME, size-sizeof(FirmwareHeader), stdio_result );

        fclose( handle );
        return ERROR_CARD_CANNOT_READ_BODY;
    }

    fclose( handle );

#if SWAP_FIRMWARE
    // we need to byte-swap the whole image, because it was already
    // proper for the card and the card swaps everything that crosses
    // the PCI interface
    //
    // here's why, tracing backwards from the PHY:
    //
    //     (1) the PHY has an 8-bit interface, so the ASIC that accepts
    //     32-bit words from the i960 must shift them when delivering
    //     them to the PHY.  it chooses to shift left, so the most
    //     significant byte of the 32-bit word is seen by the PHY first
    //
    //     (8a) what does the PCI master interface on card do?
    //
    //     (8b) what does the PCI slave interface on card do?
    //
    //     (9) the Intel Triton chipset, which controls PCI access to
    //     host memory, knows that the Intel host stores 32-bit values
    //     in memory in little-endian order, so when it served up a word
    //     to the PCI device, it placed the first byte at the low end of
    //     the word
    //
    for( count=0,swapper=(Bit32 *)start; count<size; count+=4,swapper++ )
        *swapper = ntohl( *swapper );
#endif

    static char buffer[ 1000 ];
    PentiumClock initial, now, alarm;
    double timing;
    Boot_status status;

    // suck up any console output from the ROM
    //
    // should be "Mon960\r\n"
    //
    buffer[ 0 ] = '\0';
    atm_uart_io( NULL, buffer );

    // maybe need to prod the NINDY monitor
    // so it will accept commands from us
    //
    sprintf( buffer, "\r" );
    atm_uart_io( buffer, NULL );

#if 0
    // wait for the command prompt from the NINDY monitor
    //
    do
    {
        buffer[ 0 ] = '\0';
        atm_uart_io( NULL, buffer );
    }
    while( !strstr( buffer, "=>" ) );
#endif

    // tell (the NINDY OS (that the card booted after reset))
    // to start the code we downloaded
    //
    sprintf( buffer, "go %4.4X\r", header.entry_address & 0xFFFF );
    atm_uart_io( buffer, NULL );

    // wait a few seconds for the boot status
    // to say the CP (cell processor) is running
    //
    initial.set();
    now = initial;
    alarm.put_seconds( FIRM_START_DELAY );
    alarm += now;

    while( now < alarm )
    {
        status = memmap->shared.mon960.bstat;
        if( status == cp_running )
            break;

        now.set();
    }

    if( status != cp_running )
    {
        printf( "card 0x%4.4X firmware took longer than %g seconds to run after download; last boot status was 0x%8.8lX\n", pci_handle, FIRM_START_DELAY, status );
        return ERROR_CARD_CP_NOT_RUNNING;
    }

    // tell the user how long this took
    //
    now -= initial;
    timing = now.get_seconds();
    printf( "card 0x%4.4X firmware took %g S to run\n", pci_handle, timing );

    return 0;
}



// check that the heartbeat is steadily decreasing
//
// assumes that wraparound will not occur now,
// since we just started the puppy, and it is a
// 32-bit value, and we hope it isn't decrementing
// very fast
//
// does not make any assumptions about how much the
// heartbeat changes each time, since that isn't documented
//
int AtmCard::atm_check_heartbeat( int want_increasing )
{
    PentiumClock initial, now, alarm;
    double timing;
    Bit32 prev_heartbeat = memmap->shared.queues.heartbeat;
    Bit32 heartbeat;
    unsigned count;
    const char *direction;

    if( want_increasing )
    {
        direction = "increase";
    }
    else
    {
        direction = "decrease";
    }

    for( count=0; count<3; count++ )
    {
        initial.set();
        now = initial;
        alarm.put_seconds( BEAT_CHANGE_DELAY );
        alarm += now;

        while( now<alarm )
        {
            heartbeat = memmap->shared.queues.heartbeat;
            if( want_increasing )
            {
                if( heartbeat > prev_heartbeat )
                    break;
            }
            else
            {
                if( heartbeat < prev_heartbeat )
                    break;
            }

            now.set();
        }

        if( now >= alarm )
        {
            printf( "card 0x%4.4X heartbeat took longer than %g seconds to %s\n", pci_handle, BEAT_CHANGE_DELAY, direction );
            return ERROR_CARD_HEARTBEAT_STUCK;
        }

        now -= initial;
        timing = now.get_seconds();
        printf( "card 0x%4.4X iteration %d took %g S to %s from %d to %d\n", pci_handle, count, timing, direction, prev_heartbeat, heartbeat );

        prev_heartbeat = heartbeat;
    }

    return 0;
}



// initialize the ATM SAR chip
//
int AtmCard::atm_init( void )
{
    Bit32 pointer;
    unsigned count;
    int result;
    PentiumClock initial, now, alarm;
    double timing;

    // we haven't had any interrupts from
    // this chip on this card yet
    //
    atm_interrupt_count = 0;

    // we haven't had any errors from this card yet
    //
    atm_interrupt_error = 0;

    // PBI has never run out of in FIFO or out FIFO entries
    //
    atm_out_fifo_full = 0;
    atm_in_fifo_almost_full = 0;

    // debug
    printf( "card 0x%4.4X initial boot status at log 0x%8.8lX is 0x%8.8lX\n",
        pci_handle,
        &memmap->shared.mon960.bstat,
        memmap->shared.mon960.bstat
    );
    printf( "card 0x%4.4X HCR at log 0x%8.8lX has value 0x%2.2X byte, 0x%8.8lX word\n",
        pci_handle,
        &memmap->dword_00h,
        *(Bit8 *)&memmap->dword_00h,
        memmap->dword_00h
    );

#if ATM_INIT_RESETS_SAR
    // set the boot status word to cold start, so the
    // i960 ?ROM code will do the right thing after we reset it
    //
    memmap->shared.mon960.bstat = cold_start;

    union { HostControl_write hcr_w; Bit32 hcr_w_alias; };

    // save any PCI registers that will get trashed by the reset
    //
    // ???

    // start a software reset
    //
    hcr_w_alias = 0;
    hcr_w.board_reset = TRUE;
#if 0
    memmap->dword_00h = hcr_w_alias;
#else
    memmap->hcr_w = hcr_w;
#endif

#if 0
    // wait for the NIC to finish resetting
    //
    // FORE recommends 63 PCI clocks
    //
    PentiumClock::wait( (ATM_RESET_DELAY / PCI_CLOCK_RATE) * CLOCK_RATE );
#else
    // 10 mS
    PentiumClock::wait( 0.01 * CLOCK_RATE );
#endif

    // stop resetting the card
    //
    hcr_w.board_reset = FALSE;
#if 0
    memmap->dword_00h = hcr_w_alias;
#else
    memmap->hcr_w = hcr_w;
#endif

    // restore any PCI registers that got trashed by the reset
    //
    // ???

    // wait a few seconds for the boot status word
    // to make a transition
    //
    Boot_status prev_status = cold_start;
    Boot_status status;

    initial.set();
    now = initial;
    alarm.put_seconds( ROM_START_DELAY );
    alarm += now;

    while( now<alarm )
    {
        status = memmap->shared.mon960.bstat;
        if( status != prev_status )
        {
            now -= initial;
            timing = now.get_seconds();

            // check that self test succeeded
            //
            if( status == self_test_ok )
                break;

            if( status == self_test_fail )
            {
                printf( "card 0x%4.4X firmware self test failed after %g seconds\n", pci_handle, timing );
                return ERROR_CARD_SELF_TEST_FAILED;
            }

            printf( "card 0x%4.4X warning: firmware transitioned to unknown status 0x%8.8lX after %g seconds\n", pci_handle, status, timing );
            prev_status = status;
        }

        now.set();
    }

    if( status == prev_status )
    {
        printf( "card 0x%4.4X ROM stayed in 0x%8.8lX state past %g seconds\n", pci_handle, status, ROM_START_DELAY );
        if( status == cold_start )
            return ERROR_CARD_BSTAT_STUCK;

        return ERROR_CARD_SELF_TEST_UNKNOWN;
    }

    // tell the user how long it took to succeed
    //
    printf( "card 0x%4.4X ROM took %g S to cold start and self test\n", pci_handle, timing );

    // download firmware to card and start it
    //
    result = atm_download();
    if( result )
        return result;

    // make sure that the heartbeat is decreasing
    //
    result = atm_check_heartbeat( FALSE );
    if( result )
        return result;

    // read the heap size (in bytes) from the card
    // and calculate the sizes of the 4 queues
    // that we want the card to have
    //

    // issue the initialize command to the card
    //
#endif

    // make sure that the heartbeat is increasing
    //
    result = atm_check_heartbeat( TRUE );
    if( result )
        return result;

    // give the card lots of buffers to place the first N bytes of each
    // packet into, but none for the rest of it, so only IP headers will
    // cross the PCI bus
    //
    // when I reprogram firmware, we will be loading only 2 pointers
    // into card at a time, and those will refer to large (~1MB) blocks
    // of contiguous physical memory, in which card will store multiple
    // packet headers, including ATM headers
    //

    // tell the card to activate all the VC's in VP 0 for receive
    //
    // when I reprogram the card, we will be specifying the desired
    // upper bits of the VPI and VCI (for comparison purposes) and
    // how many bits of VPI will be used for indexing (stealing from
    // VCI)
    //

    // disable per-packet interrupts; way too much host CPU load!
    //
    // when I reprogram firmware, this could be better amortized
    // (one interrupt per block of many packet headers)
    //
    union { HostInterruptMask himr; Bit32 himr_alias; };
    himr_alias = 0;
    himr.is_masked = TRUE;
    memmap->dword_01h = himr_alias;

    // allow timer interrupts
    //
    timer_interrupts_allowed = TRUE;

    return 0;
}



void AtmCard::atm_exit( void )
{
    int result;
    unsigned count;
    double timing;
    Bit8 value;
    Bit32 buffer_count, block_count;

    // ignore timer interrupts from now on
    //
    timer_interrupts_allowed = FALSE;

    // maybe suck the ATM statistics out of the card?
    //

    // tell the card to deactivate receive on all those
    // VC's we activated in atm_init() above
    //

    // disable all future card interrupts
    //
    union { HostInterruptMask himr; Bit32 himr_alias; };
    himr_alias = 0;
    himr.is_masked = TRUE;
    memmap->dword_01h = himr_alias;




    // dump our meager statistics
    //

    if( atm_out_fifo_full )
        printf( "card 0x%4.4X PBI out FIFO was full %d times during run\n", pci_handle, atm_out_fifo_full );
    else
        printf( "card 0x%4.4X PBI out FIFO was available all during run\n", pci_handle );

    if( atm_in_fifo_almost_full )
        printf( "card 0x%4.4X PBI in  FIFO was almost full %d times during run\n", pci_handle, atm_in_fifo_almost_full );
    else
        printf( "card 0x%4.4X PBI in  FIFO was available all during run\n", pci_handle );

    printf( "card 0x%4.4X ATM last interrupt status was 0x%8.8lX\n", pci_handle, atm_status_alias );
    atm_status_alias = memmap->dword_00h;
    printf( "card 0x%4.4X ATM final status 0x%8.8lX\n", pci_handle, atm_status_alias );
}



// find the FORE ATM NIC's on the PCI bus
// and initialize them
//
int pci_init( void )
{
    unsigned count;
    Bit8 value;
    Bit16 handle;
    Bit32 address;
    unsigned irq_level;
    int result;

    // check whether PCI BIOS is present
    //
    if( !pci_is_bios_present() )
    {
        printf( "Sorry, PCI BIOS is not present.\n" );
        exit( 1 );
    }

    printf( "searching for FORE PCA200E ATM network interface cards\n"
        "using vendor ID 0x%4.4X and device ID 0x%4.4X\n",
        FORE_VENDOR_ID, FORE_PCA200E_DEVICE_ID
    );

    // for all the ATM cards that are on the PCI bus...
    //
    for( count=0; count<1000; count++ )
    {
        handle = pci_get_device_handle( FORE_PCA200E_DEVICE_ID,
            FORE_VENDOR_ID, count );
        if( RETCODE_NO_MORE_PCI_DEVICES == handle )
            break;

        printf( "handle 0x%4.4X found\n", handle );

        address = pci_get_base_address( handle );
        irq_level = pci_get_irq_level( handle );

        printf( "device has physical memory address 0x%8.8lX and IRQ level %d\n", address, irq_level );

        if( count < MAX_ATM_NICS )
        {
            cards[ count ] = new AtmCard( handle, address, irq_level, count );
#if PCI_INIT_STOPS_EARLY
            count = MAX_ATM_NICS;
            break;
#endif
        }
    }

    if( count > MAX_ATM_NICS )
        printf( "warning: %d devices were found, but only %d will be used\n", count, 1 );
    else if( count < MAX_ATM_NICS )
    {
        printf( "error: only %d of expected %d devices were found!\n", count, MAX_ATM_NICS );
        exit( 1 );
    }
    else
        printf( "success!  exactly %d FORE PCA200E ATM NIC's were found\n", MAX_ATM_NICS );

    printf( "PCI configuration discovery done\n\n" );

    return 0;
}



// shut down whatever cards we were using
//
int pci_exit( void )
{
    int count;
    AtmCard *this_card;
    for( count=MAX_ATM_NICS-1; count>=0; count-- )
    {
        // we must set global pointer to nil before we
        // deallocate the card, because otherwise the next
        // timer interrupt will try to use the object after
        // it has been deallocated and possibly reused,
        // which is unsafe for MANY reasons
        //
        this_card = cards[ count ];
        cards[ count ] = NULL;
        delete this_card;

        // separate messages for one card from the other
        //
        printf( "\n" );
    }

    return 0;
}



int main( void )
{
    int result;

    // save old IRQ masks
    //
    irq = new InterruptController;

    // get the addresses and interrupts of
    // the PCI ATM cards
    //
    if( pci_init() )
        goto cleanup;

    // note: we cannot flush the level 1 cache because
    // the Borland PowerPack for DOS 32-bit protected
    // mode environment puts us in ring 3, so executing
    // the WBINVD instruction would cause a protection
    // violation
    //
    // we cannot fill the code cache because the
    // Pentium processor maintains separate data
    // and code caches, so reads of the code segment
    // would only affect the data cache
    //

    printf( "\n" );




cleanup:
    // shut down the cards
    //
    pci_exit();

    // restore old IRQ masking by destroying InterruptController
    //
    delete irq;

    // ???maybe force the CMOS clock to be read
    // into the BIOS s/w based real time clock
    //

    return result;
}


