/* conditional compile switches all go here
 */

#define FORCE_LARGE_FILES       FALSE
#define ALLOW_DISK_WRITE        TRUE
#define DEBUG_VARIANCE          FALSE
#define REPORT_OVERHEAD_STATS   FALSE




/* dostats.c */
/* compute global stats on flows */
/* jmm & kcc 29 Jul 95 */

#define MAX_PROTOCOL 255
#define MAX_PORT 65536

#if __unix__
#   include <sys/types.h>
#   include <sys/time.h>
#   include <netinet/in.h>
#else
#   include "..\wattcp\tcp.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>

#if __unix__
#   include "types.h"
#else
#   include "..\types\types.h"
#   include "..\intel\byteswap.h"
#endif

#include "queue.h"
#include "trace.h"              /* must be included before util.h */
#include "util.h"
#include "dostats.h"
#include "portstat.h"
#include "netstats.h"
#include "hoststat.h"



// none of this code gets compiled in
// if it's isn't going to be called
//
#if DO_FOREGROUND_STATS



static Bit32    flows;
static double   packetsum, bytesum, dursum;
static double   packet2sum, byte2sum, dur2sum;
static double   sumpb, sumpd, sumbd;
static int      proto_packets[MAX_PROTOCOL+1],
                proto_flows[MAX_PROTOCOL+1];
static double   proto_bytes[MAX_PROTOCOL+1];
static double   proto_duration[MAX_PROTOCOL+1];
/* arrays for variance/covar */
static double sum2_packets[MAX_PROTOCOL+1], sum2_bytes[MAX_PROTOCOL+1],
                sum2_duration[MAX_PROTOCOL+1];
static double sum_pb[MAX_PROTOCOL+1], sum_bd[MAX_PROTOCOL+1],
                sum_pd[MAX_PROTOCOL+1];




// aha!  if you don't convert n to floating point
// before squaring it, its 32 bits will overflow
// if it was greater than 64K
//
double variance(int n, double sum, double sum2)
{
#if !ALLOW_INFINITE_RESULTS
    if( n == 0 )
        return 0;
#endif

    return ((n * sum2) - (sum * sum))
    /      ( (double)n * (n>1 ? n-1 : 1) );
}



// forgets all statistics
//
int init_stats( void )
{
    Bit32 count;
    int result;

    flows = 0;
    packetsum = 0;
    bytesum = 0;
    packet2sum = 0;
    byte2sum = 0;
    dursum = 0;
    dur2sum = 0;
    sumpb = 0;
    sumpd = 0;
    sumbd = 0;

    for (count=0;count<=MAX_PROTOCOL;count++)
    {
        proto_flows[count]=0;
        proto_packets[count]=0;
        proto_bytes[count]=0;
        proto_duration[count]=0;
        sum2_packets[count]=0;
        sum2_bytes[count]=0;
        sum2_duration[count]=0;
        sum_pb[count]=0;
        sum_bd[count]=0;
        sum_pd[count]=0;
    }

    result = init_port_stats();
    if( result )
        return result;

    result = init_host_stats();
    if( result )
        return result;

    return init_net_stats();
}



/* Compute the following statistics:
 * mean and variance of packets, bytes, and duration
 * covariance
 */
void update_stats( struct hashrec *hr, float duration )
{
    u_char  proto = hr->hr_fd.fd_proto;
    u_int   packets = hr->hr_packets;
    Bit64   bytes = hr->hr_bytes;
    double packets2 = (double)packets * packets;
    double bytes2 = (double)bytes * bytes;
    double duration2 = (double)duration * duration;
    double packets_bytes = (double)packets * bytes;
    double bytes_duration = (double)bytes * duration;
    double packets_duration = (double)packets * duration;

    flows++;

    packetsum += (double)packets;
    packet2sum += packets2;

    bytesum += (double)bytes;
    byte2sum += bytes2;

    dursum += (double)duration;
    dur2sum+= duration2;

    proto_flows[proto]++;
    proto_packets[proto]+=packets;
    proto_bytes[proto]+= (double) bytes;
    proto_duration[proto]+=(double) duration;

    sum2_packets[proto]+=packets2;
    sum2_bytes[proto]+=bytes2;
    sum2_duration[proto]+=duration2;
    sum_pb[proto]+=packets_bytes;
    sum_bd[proto]+=bytes_duration;
    sum_pd[proto]+=packets_duration;
    sumpb+=packets_bytes;
    sumbd+=bytes_duration;
    sumpd+=packets_duration;

    update_port_stats( hr, proto, packets, bytes, duration, packets2, bytes2, duration2, packets_bytes, bytes_duration, packets_duration );
    update_host_stats( hr, proto, packets, bytes, duration, packets2, bytes2, duration2, packets_bytes, bytes_duration, packets_duration );
    update_net_stats( hr, packets, bytes, duration, packets2, bytes2, duration2, packets_bytes, bytes_duration, packets_duration );
}



int report_stats( tcp_Socket *sock, float istamp, float trace_duration )
{
        int result;
        int i;
        FILE *statsfile2;
        unsigned interface;
#if REPORT_OVERHEAD_STATS
        PentiumClock start, now;
        double elapsed;
        char report_oh_line[ 200 ];
        int report_oh_so_far = 0;
        int report_oh_now;
#endif

#if !ALLOW_DISK_WRITE
        fprintf( stderr, "simulated report after %g seconds to file '%s' done\n", trace_duration, filename );
        return 0;
#endif

#if REPORT_OVERHEAD_STATS
        start.set();
#endif
        sock_printf( sock, "following statistics were calculated at query time: (now)\n" );

        sock_printf (sock, "ST\t%.0f\t%.3f\t%d\t%.f\t%.lf\t%.0f\t%.0f\t%.0f\t%.0f\t"
                                "1.0\t1.0\t1.0\t1.0\t"
                                "%.3f\t%.0f\t%.0f\t"
                                "%.3f\t%.0f\t%.3f\t%.0f\t"
                                "%.3f\t%.3f\t%.3f\n",
                istamp,
                trace_duration,

                flows,
                packetsum,
                bytesum,
                dursum,

                flows/nonzero(trace_duration),
                packetsum/nonzero(trace_duration),
                bytesum/nonzero(trace_duration),

                (float) packetsum/nonzero(flows),
                (float) bytesum/nonzero(flows),
                dursum/nonzero(flows),

                (float) bytesum/nonzero(packetsum),

                variance(flows,packetsum,packet2sum),
                variance(flows,bytesum,byte2sum),
                variance(flows,dursum,dur2sum),

                sumpb,
                sumbd,
                sumpd);

        /* this is just a legend, ignored by postprocessing
         */
        sock_printf( sock, "SP [SB] lines columns:\n"
            "   Prot [sport dport] flows packets bytes dur %%f %%p %%b %%d fps\n"
            "   pps bps ppf bpf spf bpp vp vb vd copb cobd copd\n"
        );

        if (flows) {
            for (i=0;i<=MAX_PROTOCOL;i++) {
                    if (proto_flows[i])  {
                            sock_printf(sock,"SP\t%d\t%d\t%d\t%.f\t%.f\t"
                                    "%.3f\t%.3f\t%.3f\t%.3f\t"
                                    "%.3f\t%.0f\t%.0f\t"
                                    "%.3f\t%.0f\t%.3f\t%.0f\t"
                                    "%.3f\t%.3f\t%.3f\t"
                                    "%.3f\t%.3f\t%.3f\n",
                                    i,
                                    proto_flows[i],
                                    proto_packets[i],
                                    proto_bytes[i],
                                    proto_duration[i],
                                    (float) proto_flows[i]/flows,
                                    (float) proto_packets[i]/packetsum,
                                    (float) proto_bytes[i]/bytesum,         // can bytes be zero if we got a packet?
                                    proto_duration[i]/dursum,
                                    proto_flows[i]/nonzero(trace_duration),
                                    proto_packets[i]/nonzero(trace_duration),
                                    proto_bytes[i]/nonzero(trace_duration),
                                    (float) proto_packets[i]/proto_flows[i],
                                    (float) proto_bytes[i]/proto_flows[i],
                                    proto_duration[i]/proto_flows[i],
                                    (float) proto_bytes[i]/proto_packets[i],
                                    variance(proto_flows[i],proto_packets[i],sum2_packets[i]),
                                    variance(proto_flows[i],proto_bytes[i],sum2_bytes[i]),
                                    variance(proto_flows[i],proto_duration[i],sum2_duration[i]),
                                    sum_pb[i],
                                    sum_bd[i],
                                    sum_pd[i]);
                    }
            }
        }

#if REPORT_OVERHEAD_STATS
    now.set();
    elapsed = PentiumClock( now - start ).get_seconds();
    sprintf( report_oh_line+report_oh_so_far, "ST&SP:%.4f %n", elapsed, &report_oh_now );
    report_oh_so_far += report_oh_now;
    start = now;
#endif

    // give the port-specific stats a chance to talk
    //
    result = report_port_stats( sock, trace_duration, flows, packetsum, bytesum, dursum );
    if( result )
        return result;

#if REPORT_OVERHEAD_STATS
    now.set();
    elapsed = PentiumClock( now - start ).get_seconds();
    sprintf( report_oh_line+report_oh_so_far, "ports:%.4f %n", elapsed, &report_oh_now );
    report_oh_so_far += report_oh_now;
    start = now;
#endif

    // give the host-specific stats a chance to talk
    //
    result = report_host_stats( sock, trace_duration, flows, packetsum, bytesum, dursum );
    if( result )
        return result;

#if REPORT_OVERHEAD_STATS
    now.set();
    elapsed = PentiumClock( now - start ).get_seconds();
    sprintf( report_oh_line+report_oh_so_far, "hosts:%.4f %n", elapsed, &report_oh_now );
    report_oh_so_far += report_oh_now;
    start = now;
#endif

    // give the net-specific stats a chance to talk
    //
    result = report_net_stats( sock,
        istamp, trace_duration, flows,
        packetsum, bytesum, dursum,
        packet2sum, byte2sum, dur2sum,
        sumpb, sumbd, sumpd
    );
    if( result )
        return result;

#if REPORT_OVERHEAD_STATS
    now.set();
    elapsed = PentiumClock( now - start ).get_seconds();
    sprintf( report_oh_line+report_oh_so_far, "nets:%.4f\n%n", elapsed, &report_oh_now );

    fputs( report_oh_line, stderr );
    sock_puts( sock, report_oh_line );
#endif

#if DEBUG_VARIANCE
    float ftemp;

    ftemp = variance(flows,packetsum,packet2sum);
    if( ftemp < 0 )
        sock_printf( sock, "negative packet/flow variance %f caused by flows=%u packetsum=%f packet2sum=%f\n", ftemp, flows, packetsum, packet2sum );

    ftemp = variance(flows,bytesum,byte2sum);
    if( ftemp < 0 )
        sock_printf( sock, "negative byte/flow variance %f caused by flows=%u bytesum=%f byte2sum=%f\n", ftemp, flows, bytesum, byte2sum );

    ftemp = variance(flows,dursum,dur2sum);
    if( ftemp < 0 )
        sock_printf( sock, "negative duration/flow variance %f caused by flows=%u dursum=%f dur2sum=%f\n", ftemp, flows, dursum, dur2sum );
#endif

#if FORCE_LARGE_FILES
    // append 100K of trailing gunk to the text report
    // to see if capture and/or flows and/or TCP breaks
    //
    for( i=0; i<2000; i++ )

        // 50 bytes each line
        //
        sock_printf( sock, "123456789012345678901234567890123456789012345678\n" );
#endif

    // ees no my yob to close socket!
    //

    return 0;
}



#endif  // DO_FOREGROUND_STATS
