/* conditional compile switches all go here
 */

#define BYTESWAP_FIELDS_EARLY   FALSE       /* must match fs2flows.c */
#define ALWAYS_DUMP_HASH_INFO   FALSE
#define FOLD_BEFORE_MODULUS     TRUE
#define FOLD_RANDOMIZES         TRUE
#define FOLD_KEEPS_16_BITS      TRUE
#define DEBUG_CHAIN_LENGTH      FALSE       // 12/24/96 deepest was ~4 at WTN
#define DEBUG_HASHINDEX         FALSE
#define DEBUG_WORKING_SET       TRUE
#define DEBUG                   FALSE
#define SORT_HOST_PRINTOUT      FALSE
#define TALK_WHEN_ALLOC_FAILS   FALSE       // for debugging when overflow occurs only!  too slow
#define ENABLE_THRESHOLDS       TRUE
#define RESTRICT_PROTOCOLS      TRUE




#if __unix__
#   include <sys/types.h>
#   include <sys/time.h>
#   include <netinet/in.h>
#else
#   include "..\wattcp\tcp.h"
#endif
#include <stdio.h>
#include <stdlib.h>             /* errno */
#include <fcntl.h>

#if __unix__
#   include "types.h"
#else
#   include "..\types\types.h"
#   include "..\intel\byteswap.h"
#endif

#include "queue.h"
#include "trace.h"              /* must be included before util.h */
#include "util.h"
#include "dostats.h"            /* nonzero() variance() */
#include "hoststat.h"
#include "flat.h"               /* flat_alloc() */



// none of this code gets compiled in
// if it's isn't going to be called
//
#if DO_FOREGROUND_STATS



static float    cutoffp, cutoffb, cutofff;
static int      hosttriplecount;

#define N_HOSTTRIPLES    65536

#if !FOLD_BEFORE_MODULUS
#   define fold(x) (x)
#   define modulus(x) ((x) % N_HOSTTRIPLES)
#else
#if FOLD_RANDOMIZES
#   if BIT64_AVAILABLE
#       define fold32(x) (((x)>>32) ^ (x))
#       define modulus(x) ((x) & 65535)
#   else
#       define fold32(x) (x)
#       define modulus(x) (x)
#   endif
#   define fold(x) ((fold32(x)*1664525U + 1013904223) >> 16)
#else
#if FOLD_KEEPS_16_BITS
#   define fold(x) (((x)>>16) ^ (x))
#   define modulus(x) ((x) & 65535)
#else
#   define fold(x) (((x)>>24) ^ ((x)>>16) ^ ((x)>>8) ^ (x))
#   define modulus(x) ((x) & 255)
#endif
#endif
#endif

/* generate a 32-bit hash key
 *
 * make sure this uses unsigned shifting
 * by explicitly promoting to Bit32
 */
#define hosthash(ip,proto,port) modulus( fold( \
                  ip ^ (((Bit32)proto << 16) | port) \
                  ) )

#define HOSTBUCKET_POOL_SIZE (N_HOSTTRIPLES * sizeof(struct hosthashrec *))
static struct hosthashrec **hosttab = NULL;
static struct hosthashrec *hosthash_free = NULL;
static struct hosthashrec *hosthash_pool = NULL;
u_int hosthash_alloc_failures;

// needed 8.9K at WTN 8:22PM 12/23/96 so alloc 18K
//
// needed >18K at WTN 2/12/97 so increase to 27K
//
//#define HOSTHASH_RECORDS 18000L
#define HOSTHASH_RECORDS 27000L
//#define HOSTHASH_RECORDS 100000L
//#define HOSTHASH_RECORDS 500000L      // not enough memory left over for packet buffering!

#define HOSTHASH_POOL_SIZE (HOSTHASH_RECORDS * sizeof(struct hosthashrec))

#if DEBUG_CHAIN_LENGTH
#   define HOSTCHAIN_POOL_SIZE (N_HOSTTRIPLES * sizeof(int))
    static int *chainlen2 = NULL;
    static int longest_chain2;
#endif

#if DEBUG_WORKING_SET
    static int total_level2_hosthashrecs;
#endif

#if DEBUG_HASHINDEX
    static u_int hashindex_trace = -1;
#endif



#if SORT_HOST_PRINTOUT
#define HOSTPRINT_POOL_SIZE (HOSTHASH_RECORDS * sizeof(struct hosthashrec *))
static Bit32 hostcount;
static struct hosthashrec **hostprintout = NULL;

static int hostcmp( void *a, void *b )
{
    struct hosthashrec ** hrp1 = (struct hosthashrec**) a;
    struct hosthashrec ** hrp2 = (struct hosthashrec**) b;
    return ((*hrp1)->packets - (*hrp2)->packets);
}
#endif



/*
 * new_host_record
 *      establish a new host triple
 *      1> insert into hash table [maintained as stack]
 */
static
struct hosthashrec *
new_host_record( u_int ip, u_char proto, u_short port,
    u_int flows, u_int packets, Bit64 bytes, float duration,
    double p2sum, double b2sum, double d2sum,
    double pbsum, double bdsum, double pdsum
)
{
    struct hosthashrec *pr;

    if ((pr = hosthash_free) == NULL) {
#if TALK_WHEN_ALLOC_FAILS
        fprintf(stderr, "unable to alloc new hosthashrec\n");
#endif
        hosthash_alloc_failures++;
        return 0;
    }
    hosthash_free = pr->next;

    pr->ip=ip;
    pr->proto=proto;
    pr->port=port;
    pr->flows=flows;
    pr->packets=packets;
    pr->bytes=bytes;
    pr->duration=duration;
    pr->p2sum=p2sum;
    pr->b2sum=b2sum;
    pr->d2sum=d2sum;
    pr->pbsum=pbsum;
    pr->bdsum=bdsum;
    pr->pdsum=pdsum;
    pr->next=NULL;
    return(pr);
}



static void delete_host_record( struct hosthashrec *pr )
{
    /* put this puppy on the free list */
    pr->next = hosthash_free;
    hosthash_free = pr;
}


/*
 * diagnostic
 */
static void print_host_record( struct hosthashrec *pr )
{
    char temp[20];

    fprintf(stderr, "p: %d ", pr->packets);
    fprintf(stderr, "b: %.f ", double(pr->bytes));
    fprintf(stderr, "sdd hash: %s %d %d %f %d\n",
        inet_ntoa(temp,pr->ip), pr->proto, pr->port,
        pr->duration, hosthash(pr->ip,pr->proto,pr->port));
}



// forgets all statistics
//
int init_host_stats( void )
{
    Bit32 count;
    int result;

#if ENABLE_THRESHOLDS
    cutoffp = .1;           // 10%    of packets
    cutoffb = .0005;        //  0.05% of bytes
    cutofff = .1;           // 10%    of flows
#else
    cutoffp = 0.0;
    cutoffb = 0.0;
    cutofff = 0.0;
#endif

    hosttriplecount = 0;

#if DEBUG_CHAIN_LENGTH
    longest_chain2 = 0;
#endif

#if DEBUG_WORKING_SET
    total_level2_hosthashrecs = 0;
#endif

    // if this is the first clear (really initialization)
    //
    if( hosthash_pool == NULL )
    {
        fprintf( stderr, "host hash records are %d bytes each\n", sizeof(struct hosthashrec) );

#if DEBUG_CHAIN_LENGTH
        /* allocate chain length counters for each bucket in hash table */
        chainlen2 = (int *) flat_alloc( HOSTCHAIN_POOL_SIZE );
        if( chainlen2 == NULL )
        {
            printf( "error %d allocating %d bytes (%d integers) for host chain length counters\n", errno, HOSTCHAIN_POOL_SIZE, N_HOSTTRIPLES );
            return 1;
        }
#endif

        /* allocate the buckets for the secondary hash table */
        hosttab = (struct hosthashrec **) flat_alloc( HOSTBUCKET_POOL_SIZE );
        if( hosttab == NULL )
        {
            printf( "error %d allocating %ld bytes (%ld pointers) for host hash buckets\n", errno, HOSTBUCKET_POOL_SIZE, N_HOSTTRIPLES );
            return 1;
        }

#if SORT_HOST_PRINTOUT
        /* allocate pointers to sort during report */
        hostprintout = (struct hosthashrec **) flat_alloc( HOSTPRINT_POOL_SIZE );
        if( hostprintout == NULL )
        {
            printf( "error %d allocating %ld bytes (%ld pointers) for host printout sorting\n", errno, HOSTPRINT_POOL_SIZE, HOSTHASH_RECORDS );
            return 1;
        }
#endif

        /* allocate a pool of hash records */
        hosthash_pool = (struct hosthashrec *) flat_alloc( HOSTHASH_POOL_SIZE );
        if( hosthash_pool == NULL )
        {
            printf( "error %d allocating %ld bytes (%ld records) for host hash record pool\n", errno, HOSTHASH_POOL_SIZE, HOSTHASH_RECORDS );
            return 1;
        }

        /* put all the hash records on the singly-linked free list */
        hosthash_free = hosthash_pool;
        for( count=0; count<HOSTHASH_RECORDS-1; count++ )
        {
            hosthash_pool[ count ].next = hosthash_pool + count + 1;
        }
        hosthash_pool[count].next = NULL;
    }
    hosthash_alloc_failures = 0;

    for( count=0; count<N_HOSTTRIPLES; count++ )
    {
        // level2 hashtable is empty
        //
        hosttab[count] = 0;

#if SORT_HOST_PRINTOUT
        // no host triples are ready to print
        //
        hostprintout[ count ] = 0;
#endif

#if DEBUG_CHAIN_LENGTH
        /* initialize chain length of all hashtable buckets to zero
         */
        chainlen2[ count ] = 0;
#endif
    }

    return 0;
}



void update_host_stats( struct hashrec *hr, u_char proto,
    u_int packets, Bit64 bytes, float duration,
    double packets2, double bytes2, double duration2,
    double packets_bytes, double bytes_duration, double packets_duration
)
{
        unsigned hosthashindex;
        int cmp;
        struct hosthashrec *pp, *hosth, *newpp;
        char temp[ 20 ];        // for inet_netoa()
#if BYTESWAP_FIELDS_EARLY
        u_int     ip = hr->hr_fd.fd_src.s_addr;
        u_short port = hr->hr_fd.fd_sport;
#else
        u_int     ip = ntohl( hr->hr_fd.fd_src.s_addr );
        u_short port = ntohs( hr->hr_fd.fd_sport );
#endif

#if RESTRICT_PROTOCOLS
        /* discard flows which are uninteresting
         */
        if( !(proto==IPPROTO_TCP && port==80)       // want web server
        &&  !(proto==IPPROTO_TCP && port==20)       // want FTP control
        &&  !(proto==IPPROTO_TCP && port==21) )     // want FTP data
        {
            return;
        }
#endif

        hosthashindex = hosthash(ip,proto,port);
        pp = hosttab[hosthashindex];
        /* look for match or collision on hash chain */
        hosth=pp;
        cmp=0;
        while (hosth!=NULL && cmp==0)
        {
                cmp = (ip==hosth->ip && proto==hosth->proto && port==hosth->port);
                if (cmp) {
                        /* update all the stats in hosthashrec */
                        hosth->flows++;
                        hosth->packets+=packets;
                        hosth->bytes+=bytes;
                        hosth->duration+=duration;
                        hosth->p2sum+=packets2;
                        hosth->b2sum+=bytes2;
                        hosth->d2sum+=duration2;
                        hosth->pbsum+=packets_bytes;
                        hosth->bdsum+=bytes_duration;
                        hosth->pdsum+=packets_duration;
#if DEBUG
                        /* for debugging */
                        printf("match\t");
                        print_host_record(hosth);
#endif
                }
                hosth=hosth->next;
        };

        /* didn't find it anywhere, must be a new record */
        if (cmp==0)
        {
            hosttriplecount++;
            newpp = new_host_record( ip, proto, port, 1, packets,
                bytes, duration, packets2, bytes2, duration2, packets_bytes,
                bytes_duration, packets_duration );
            if( newpp )
            {
                newpp->next = pp;
                hosttab[hosthashindex] = newpp;
#if DEBUG_WORKING_SET
                /* increment count of level2 hash records
                 */
                total_level2_hosthashrecs++;
#endif

#if DEBUG_CHAIN_LENGTH
                /*
                 * increment chain length for this hash index
                 */
                chainlen2[hosthashindex]++;

#if DEBUG_HASHINDEX
                if( hosthashindex == hashindex_trace )
                    fprintf( stderr, "at packet %d hash index 0x%8.8lX chain length increased to %d\n", total_packets, hosthashindex, chainlen2[hosthashindex] );
#endif

                /*
                 * if this is the longest chain we have ever seen,
                 * update our global, and notify the user
                 *
                 * notifications should be asymptotically decreasing
                 */
                if( longest_chain2 < chainlen2[hosthashindex] )
                {
                    longest_chain2 = chainlen2[hosthashindex];
                    fprintf( stderr, "hash index 0x%8.8lX is now longest host chain 2 at %d\n", hosthashindex, longest_chain2 );
                    fprintf( stderr, "calced from ip=%s proto=%d port=%d\n", inet_ntoa(temp,ip), proto, port );
                }
#endif

/* for debugging */
#if DEBUG
                printf("new\t");
                print_host_record(newpp);
#endif
            }   // end if could alloc hosthashrec
        }   // end if needed a new hosthashrec
}



/* "SHB ip proto port flows packets bytes dur %%f %%p %%b %%d fps "
   "pps bps ppf bpf spf bpp vp vb vd copb cobd copd \n");
*/

static void print_host_stats( tcp_Socket *sock, float trace_duration,
    int total_flows, struct hosthashrec *hosth, double packetsum,
    double bytesum, double dursum )
{
    char temp[20];

    sock_printf(
        sock,
        "SHB\t%s\t%d\t%d\t%d\t%d\t%.0f\t%.3f\t"
            "%.3f\t%.3f\t%.3f\t%.3f\t"
            "%.3f\t%.0f\t%.0f\t"
            "%.3f\t%.0f\t%.3f\t%.0f\t"
            "%.3f\t%.3f\t%.3f\t"
            "%.3f\t%.3f\t%.3f\n",

        inet_ntoa(temp,hosth->ip),
        hosth->proto,
        hosth->port,

        hosth->flows,
        hosth->packets,
        (double) hosth->bytes,
        hosth->duration,

        (float) hosth->flows/total_flows,
        (float) hosth->packets/packetsum,
        (float) hosth->bytes/bytesum,
        (float) hosth->duration/dursum,

        hosth->flows/nonzero(trace_duration),
        hosth->packets/nonzero(trace_duration),
        hosth->bytes/nonzero(trace_duration),

        (float) hosth->packets/hosth->flows,
        (float) hosth->bytes/hosth->flows,
        (float) hosth->duration/hosth->flows,

        (float) hosth->bytes/hosth->packets,

        variance(hosth->flows,hosth->packets,hosth->p2sum),
        variance(hosth->flows,hosth->bytes,hosth->b2sum),
        variance(hosth->flows,hosth->duration,hosth->d2sum),

        (float) hosth->pbsum,
        (float) hosth->bdsum,
        (float) hosth->pdsum                // 24th field
    );
}



int report_host_stats( tcp_Socket *sock, float trace_duration, u_int flows,
    double packetsum, double bytesum, double dursum )
{
        int i;
        unsigned hosthashindex;
        struct hosthashrec *hosth, *next;
        Bit64 cutoffb_bit64;

#if DEBUG_WORKING_SET
        fprintf( stderr, "%d level 2 host hash records created\n", total_level2_hosthashrecs );
#endif

        /* go through hosttab and print  now */
        cutoffp *= packetsum;
        cutoffb *= bytesum;
        cutoffb_bit64 = cutoffb;
        cutofff *= flows;

        for (i=0;i<N_HOSTTRIPLES;i++) {
            for (hosth=hosttab[i]; hosth!=NULL; hosth=next) {
                next = hosth->next;

                if ((hosth->packets > cutoffp) ||
                    (hosth->bytes > cutoffb_bit64) ||
                    (hosth->flows > cutofff))
                {
#if SORT_HOST_PRINTOUT
                    netcount++;
                    hostprintout[hostcount]=hosth;
#else
                    print_host_stats(sock, trace_duration, flows, hosth, packetsum, bytesum, dursum);
#endif
                }
#if SORT_HOST_PRINTOUT
                /* only want to delete unselected host triples now;
                 * selected ones will be deleted after sort/print
                 */
                else
#else
                /* want to delete both printed and unprinted host triples
                 */
#endif
                    delete_host_record( hosth );
            }
        }

#if SORT_HOST_PRINTOUT
        /* sort from hosthle array 80 each */
        qsort( hostprintout, hostcount, sizeof(struct hosthashrec *), &hostcmp );
        for( i=0; i<hostcount; i++)
        {
            print_host_stats( sock, trace_duration, hostprintout[i],packetsum, bytesum, dursum );
            delete_host_record( hostprintout[i] );
        }
#endif

    return 0;
}



#endif  // DO_FOREGROUND_STATS
