#ifndef _NETINET_TCP_H
#define _NETINET_TCP_H

typedef u_int   tcp_seq;
/*
 * TCP header.
 * Per RFC 793, September, 1981.
 */
struct tcphdr {
    u_short th_sport;       /* source port */
    u_short th_dport;       /* destination port */
    tcp_seq th_seq;         /* sequence number */
    tcp_seq th_ack;         /* acknowledgement number */
    u_char  th_x2:4,        /* (unused) */
            th_off:4;       /* data offset */
    u_char  th_flags;
#define TH_FIN  0x01
#define TH_SYN  0x02
#define TH_RST  0x04
#define TH_PUSH 0x08
#define TH_ACK  0x10
#define TH_URG  0x20
    u_short th_win;         /* window */
    u_short th_sum;         /* checksum */
    u_short th_urp;         /* urgent pointer */
};

#define TCPOPT_EOL  0
#define TCPOPT_NOP  1
#define TCPOPT_MAXSEG   2
#define TCPOPT_WINDOW   3           /* rfc 1323 window scale option */

/*
 * Default maximum segment size for TCP.
 * For RFC1122 MUST conformance, this needs to be 536.
 */
#define TCP_MSS 536

#define TCP_MAXWIN  65535       /* largest value for (unscaled) window */

#define TCP_MAX_WINSHIFT      14    /* maximum window shift */

/*
 * User options (used with set/getockopt).
 */
//#define TCP_NODELAY 0x01    /* don't delay send to coalesce packets */
#define TCP_MAXSEG  0x02    /* maximum segment size */
/*
 * Number of retransmissions before dropping connection. Larger than
 * TCP_MAXRXTSHIFT or -1 for infinity
 */
#define TCP_RPTR2RXT    0x03    /* set repeat count for R2 RXT timer */

#endif
