// force these variables so this executable will always
// operate on an IP header trace, regardless how CAPTURE.EXE
// is compiled (as determined by CONSTANT.H)
//
// this is only safe to do if this is the only source file
// in this executable which cares about tracerec structure
//
#define CAPTURE_TO_DISK             FALSE
#define CELLS_PER_BUFFER            1

#include <mem.h>                        // memset()
#include <stdio.h>                      // fopen()
#include <errno.h>                      // errno
#include <conio.h>                      // kbhit()
#include "trace.h"
#include "..\intel\byteswap.h"          // htons() htonl()



// conditional compilation switches
//
#define ALLOW_KEYPRESS_TO_HALT          TRUE
#define PRINT_LEADING_ZEROES            FALSE
#define PRINT_LENGTHS_WITH_NO_PACKETS   FALSE
#define TRACE_ALREADY_SWAPPED           TRUE
#define FIX_CARD_INTERRUPT_RACE_COND    FALSE
#define WANT_TIMES                      TRUE



// global variables
//
struct traceblock buffer;
Bit32 ip_length_usage[ IP_MAXPACKET + 1 ];



int main( int argc, char *argv[] )
{
    const char *program = argv[ 0 ];
    const char *infile  = argv[ 1 ];
#if WANT_TIMES
    const char *outfile = argv[ 2 ];
    const char *cardstr = argv[ 3 ];
#else
    const char *cardstr = argv[ 2 ];
#endif
    FILE *in_handle;
    FILE *out_handle;
    int result;
    Bit32 used;
    Bit32 count;
    Bit32 block_count = 0;
    Bit32 unknown_llc_packets = 0;
    Bit32 unknown_llc_and_non_ip_ethertype_packets = 0;
    Bit32 unknown_oui_packets = 0;
    Bit32 unknown_oui_and_non_ip_ethertype_packets = 0;
    Bit32 non_ip_ethertype_packets = 0;
    Bit32 valid_packets = 0;
    Bit32 ip_optioned_packets = 0;
    Bit32 ip_unknown_flagged_packets = 0;
    Bit32 ip_df_flagged_packets = 0;
    Bit32 ip_whole_packets = 0;
    Bit32 first_ip_fragments = 0;
    Bit32 non_first_ip_fragments = 0;
    Bit32 wrong_cardnum_packets = 0;
    Bit32 when_temp;
    struct tracerec *packet;
    Bit16 cardnum_wanted;
#if FIX_CARD_INTERRUPT_RACE_COND
    Bit32 fixed_timestamp_packets = 0;      // included in valid_packets
    CardClock prev_when;
    CardClock fixup_when;
    fixup_when.low = 65536;
    fixup_when.high = 0;
#endif

    fprintf( stdout, "each record is %d bytes\n", sizeof(struct tracerec) );
    fprintf( stdout, "ICMP alone accounts for %d bytes\n", sizeof(struct icmp) );
    fprintf( stdout, "%d records in a block\n\n", COMPL_RING_ENTRIES );

#if WANT_TIMES
    if( argc < 3 )
#else
    if( argc < 2 )
#endif
    {
        fprintf( stderr, "extracts the timestamp and IP length in ASCII hex from an OC3MON trace file\n" );
        fprintf( stderr, "and counts usage of each IP length\n\n" );
#if WANT_TIMES
        fprintf( stderr, "usage: %s infile outfile {hex-cardnum}\n", program );
#else
        fprintf( stderr, "usage: %s infile {hex-cardnum}\n", program );
#endif
        return 1;
    }

    in_handle = fopen( infile, "rb" );
    if( in_handle == NULL )
    {
        fprintf( stderr, "error %d opening input file\n", errno );
        return 2;
    }

#if WANT_TIMES
    out_handle = fopen( outfile, "w" );
    if( out_handle == NULL )
    {
        fprintf( stderr, "error %d opening output file\n", errno );
        return 3;
    }

    if( argc >= 4 )
#else
    if( argc >= 3 )
#endif
    {
        result = sscanf( cardstr, "%hX", &cardnum_wanted );
        if( result != 1 )
        {
            fprintf( stderr, "unable to convert cardnum '%s' to integer\n", cardstr );
            return 4;
        }
    }
    else
        cardnum_wanted = 0;

    // initialize packet length histogram to empty
    //
    memset( ip_length_usage, 0, sizeof(ip_length_usage) );

    while( !feof( in_handle ) && !ferror( in_handle )
#if ALLOW_KEYPRESS_TO_HALT
        && !kbhit()
#endif
    )
    {
        block_count++;
        fprintf( stderr, "%lu ", block_count*sizeof(struct traceblock) );

        result = fread( &buffer, sizeof(buffer), 1, in_handle );
        if( result != 1 )
        {
            if( !feof( in_handle ) )
                fprintf( stderr, "error %d reading input file\n", errno );

            break;
        }

#if HOST_IS_BIG_ENDIAN != TRACE_ALREADY_SWAPPED
        // swap the usage count and card number
        //
        buffer.used = htonl( buffer.used );
        buffer.cardnum = htons( buffer.cardnum );
#endif

        // must remember usage count
        //
        // at this point it is in host byte order
        //
        used = buffer.used;

        // check for out-of-bound record count
        //
        if( used > COMPL_RING_ENTRIES )
        {
            fprintf( stderr, "error: usage count of block %u is %u, but max is %u!\n", block_count, used, COMPL_RING_ENTRIES );
            break;
        }

        if( cardnum_wanted!=0 && buffer.cardnum!=cardnum_wanted )
        {
            fprintf( stderr, "ignoring cardnum 0x%.4X\n", buffer.cardnum );
            wrong_cardnum_packets += used;
            continue;
        }

        // for each record in the buffer...
        //
        for( count=0; count<used; count++ )
        {
            // shortcut for code that follows
            //
            packet = &buffer.data[ count ];

            // if LLC is bad...
            //
            if( packet->tr_llc.lsap != LSAP_SNAP
            ||  packet->tr_llc.ctl != CTL_SNAP )
            {
                unknown_llc_packets++;
                if( packet->tr_snap.ethertype != htons( ETHERTYPE_IP ) )
                    unknown_llc_and_non_ip_ethertype_packets++;
                continue;
            }

            // if OUI is bad...
            //
            if( packet->tr_snap.oui[0] != 0
            ||  packet->tr_snap.oui[1] != 0
            ||  packet->tr_snap.oui[2] != 0 )
            {
                unknown_oui_packets++;
                if( packet->tr_snap.ethertype != htons( ETHERTYPE_IP ) )
                    unknown_oui_and_non_ip_ethertype_packets++;
                continue;
            }

            // if ethertype is not IP...
            //
            if( packet->tr_snap.ethertype != htons( ETHERTYPE_IP ) )
            {
                non_ip_ethertype_packets++;
                continue;
            }

            // packet is good as far as we are concerned
            //
            // note that flows analysis will not deal with
            // packets with IP options or which are not first
            // of fragment because TCP header has shifted
            // out of the cell, and it doesn't want to bother
            // re-laying out the structure to get ICMP payload
            //
            // also note that flows analysis incorrectly calculates
            // length for IP fragmented packets, since IP header's
            // length in first fragment includes only that fragment,
            // and it doesn't count length of non-first fragments
            //
            valid_packets++;

            // if this packet used IP options...
            //
            if( packet->tr_ip.ip_hl != (sizeof(ip) >> 2) )
            {
                ip_optioned_packets++;
            }

#if HOST_IS_BIG_ENDIAN != TRACE_ALREADY_SWAPPED
            // swap the bytes and words of the timestamp
            // so we can understand them (they are in
            // network byte order on disk)
            //
            when_temp = ntohl( packet->when.low );
            packet->when.low  = ntohl( packet->when.high );
            packet->when.high = when_temp;

            // swap the bytes of the length in the IP header
            //
            packet->tr_ip.ip_len = ntohs( packet->tr_ip.ip_len );

            // swap the bytes of the IP fragment offset
            //
            packet->tr_offset = ntohs( packet->tr_offset );
#endif

            // note that flows analysis would not deal with
            // fragments because TCP header has been shifted
            // and it doesn't want to recalculate field
            // locations
            //

            // if this is a followon fragment...
            //
            if( (packet->tr_offset & IP_OFF_MASK) != 0 )
            {
                non_first_ip_fragments++;
            }

            // otherwise, if more fragments will follow...
            //
            else if( packet->tr_offset & IP_MF )
            {
                first_ip_fragments++;
            }

            // otherwise this must be all of an unfragmented packet
            //
            else
            {
                // if this packet used other IP flags...
                //
                if( packet->tr_offset & IP_UF )
                    ip_unknown_flagged_packets++;
                else if( packet->tr_offset & IP_DF )
                    ip_df_flagged_packets++;
                else
                    ip_whole_packets++;
            }

            // update the count of packets with this length
            //
            ip_length_usage[ packet->tr_ip.ip_len ]++;

#if FIX_CARD_INTERRUPT_RACE_COND
            // if race between the card's recording hardware and software
            // portions of timestamp caused timestamp to either get smaller
            // or increase by more than hardware could do by itself...
            //
            // warning: that check for too large an increase is unreliable
            // in an idle network like a test link, because macroscopic
            // delays could indeed have occurred between some packets!
            //
            if( packet->when<prev_when
            || (packet->when>=prev_when+fixup_when
                && ((packet->when.low & 0xFF00) == 0xFF00) )
            )
            {
                packet->when -= fixup_when;
                fixed_timestamp_packets++;
            }

            prev_when = packet->when;
#endif

#if WANT_TIMES
            // print the timestamp and length in hex to stdout
            //
#if PRINT_LEADING_ZEROES
            fprintf( out_handle, "%.8lX%.8lX %.4X\n", packet->when.high, packet->when.low, packet->tr_ip.ip_len );
#else
            if( packet->when.high > 0 )
                fprintf( out_handle, "%lX%.8lX %X\n", packet->when.high, packet->when.low, packet->tr_ip.ip_len );
            else
                fprintf( out_handle, "%lX %X\n", packet->when.low, packet->tr_ip.ip_len );
#endif
#endif
        }
    }
    fprintf( stderr, "\n" );

#if ALLOW_KEYPRESS_TO_HALT
    if( kbhit() )
    {
        getch();
        fprintf( stderr, "halted by keypress!\n" );
    }
#endif

    // show statistics
    //
    fprintf( stdout, "%lu blocks examined\n", block_count );
    fprintf( stdout, "%lu packets with unknown LLC (not 0xAA 0xAA 0x03) found\n", unknown_llc_packets );
    fprintf( stdout, "%lu of unknown LLC packets also had non IP ethertype\n", unknown_llc_and_non_ip_ethertype_packets );
    fprintf( stdout, "%lu packets with unknown SNAP OUI (not 0x00 0x00 0x00) found\n", unknown_oui_packets );
    fprintf( stdout, "%lu of unknown SNAP OUI packets also had non IP ethertype\n", unknown_oui_and_non_ip_ethertype_packets );
    fprintf( stdout, "%lu packets with non IP (not 0x0800) ethertype found\n", non_ip_ethertype_packets );
    if( cardnum_wanted != 0 )
        fprintf( stdout, "%lu packets without cardnum 0x%.4X found\n", wrong_cardnum_packets, cardnum_wanted );
    fprintf( stdout, "%lu valid packets found\n", valid_packets );
    fprintf( stdout, "%lu of valid packets used IP options\n", ip_optioned_packets );
    fprintf( stdout, "%lu of valid packets were whole and unflagged\n", ip_whole_packets );
    fprintf( stdout, "%lu of valid packets were whole and used unknown IP flag\n", ip_unknown_flagged_packets );
    fprintf( stdout, "%lu of valid packets were whole and asked not to be fragmented\n", ip_df_flagged_packets );
    fprintf( stdout, "%lu of valid packets were first of fragment\n", first_ip_fragments );
    fprintf( stdout, "%lu of valid packets were not first of fragment\n", non_first_ip_fragments );
#if FIX_CARD_INTERRUPT_RACE_COND
    fprintf( stdout, "%lu of valid packets needed timestamp fixing\n", fixed_timestamp_packets );
#endif

    printf( "number (in decimal) of packets for each IP length (in decimal):\n" );
    for( count=0; count<=IP_MAXPACKET; count++ )    // assumes count has >16 bits
    {
#if !PRINT_LENGTHS_WITH_NO_PACKETS
        if( ip_length_usage[count] > 0 )
#endif
            fprintf( stdout, "%u %lu\n", count, ip_length_usage[count] );
    }

#if WANT_TIMES
    fclose( out_handle );
#endif
    fclose( in_handle );

    // return successfully
    //
    return 0;
}

