#include <dos.h>
#include <conio.h>
#include <alloc.h>
#include <ctype.h>
#include <errno.h>

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>

#include "..\snmplib\ausnmp.h"
#include "..\wattcp\tcp.h"
#include "..\types\types.h"
#include "..\wattcp\tcp.h"
#include "..\wattcp\pktdrv.h"

/* Declared in wattcp\src\pcpkt.c .. */

extern unsigned char eth_addr[6];
extern unsigned long npackets,nbytes;

#define MAXPKTLEN  1526

udp_Socket udpsock;
static udp_Socket *s;
int status;  /* For socket */
word _pktdevclass = 1;      /* Ethernet = 1, SLIP = 6 */

/* forward declarations */
void accept_pkt(unsigned int len, unsigned char far *buff, 
   unsigned int i_face, unsigned int lah_len);
int our_packet(unsigned char far *buff);

void receive(void);
void snmp_read(int length);

#define TRACE_SZ 5000
struct p_t {
   Bit32 us;  /* Inter-arrival time, microseconds */
   Bit16 sz;  /* Packet size, bytes */
   };
struct p_t p_table[TRACE_SZ];   
long n_p = -1;

double uptime(void);  /* In time_us.cpp */
double last_time;

void accept_pkt(unsigned int len, unsigned char far *buff, 
   unsigned int i_face, unsigned int lah_len)
{
   double now = uptime();
   Bit32 i = (now-last_time)*1000000;
   if (n_p != TRACE_SZ) {
      if (n_p >= 0) {
         p_table[n_p].us = i;
         p_table[n_p].sz = len;
         }
      ++n_p;
      }
   last_time = now;
   }


#define SNAPSIZE  54  /* Matches meter_ux.c */

int our_packet(unsigned char far *buff)
{
   if (buff[0] == 0xFF) {  /* Broadcast */
      if (buff[13] == 0x06 && buff[12] == 0x08)  /* ARP */
	 return 1;
      }
   else if (buff[5] == eth_addr[5] && buff[4] == eth_addr[4] &&
	 buff[3] == eth_addr[3] && buff[2] == eth_addr[2] &&
	 buff[1] == eth_addr[1] && buff[0] == eth_addr[0]) {
      if (buff[12] == 0x08 &&
	    (buff[13] == 0x00 || buff[13] == 0x06 )) {  /* IP or ARP */
	 return 1;
	 }
      }
   return 0;  /* Only want our IP & ARP pkts */
   }

Bit8 *pkt_rcv_call1(Bit16 len, Bit16 lah_len, Bit8 *buff, Bit16 if_nbr)
{
   int if_type = en_table[if_nbr].type;
   if (if_type & EN_HP) {
      if (if_type & EN_MT)
	 accept_pkt(len, buff, if_nbr, lah_len);  /* Pass attribute info to meter buffer */
      if (if_type & EN_IP) {
	 if (our_packet(buff)) return get_pktbuf(len);
	 }
      return (unsigned char far *)NULL;  /* Don't bother with call2 */
      }
   else  /* Can't see packet contents until call2 */
      return get_pktbuf(len);
   }

void pkt_rcv_call2(Bit16 len, Bit8 *buff, Bit16 if_nbr)
{
   int if_type = en_table[if_nbr].type;
   struct pbuf *bp;
   int ip_pkt = 0;
   if (if_type & EN_HP)
      ip_pkt = 1;  /* IP pkt ready for tcp/ip */
   else {
      if (if_type & EN_MT) {
         accept_pkt(len, buff, if_nbr, 0);  /* Pass attribute info to meter buffer */
         if (if_type & EN_IP)
            ip_pkt = our_packet(buff) ? 1 : 0;
         else ip_pkt = 0;
         }
      else if (if_type & EN_IP)
	 ip_pkt = 1;  /* Unicast/broadcast mode */
      else ip_pkt = 0;  /* Not our packet, free the buffer */
      }
   bp = (struct pbuf *)(buff - BUF_HDR_LEN);
   bp->state = 0;
   enqueue(&freeq, bp);
   return;
   }

int snmp_dump_packet = 0;

int main(int argc, char *argv[])
{
   int arg, a, c;
   char cbuf[20], *ap;
   int n_interfaces, IP_int;

   n_interfaces = 0;  IP_int = -1;
   for (c = 0, arg = 1; arg < argc; ++arg) {
      if (argv[arg][0] == '-') {
         ap = argv[arg]+2;
         switch (a = argv[arg][1]) {
	 case 'I':  /* IP only - no metering */
	    IP_int = n_interfaces;
	 case 'i':  /* Just plain metering */
	 case 'h':  /* Metering with high-performance driver */
	    if (*ap == NULL) ap = argv[++arg];
            if (arg == argc) c = 0;  /* No more args */
            else c = isdigit(*ap) ? atoi(ap) : 0;  /* -h means -h0 */
	    en_table[n_interfaces].int_nbr = c;
	    if (a != 'I') en_table[n_interfaces].type =
	       a == 'h' ? EN_MT|EN_HP : EN_MT;
	    ++n_interfaces;
	    break;
	 default:
	    printf("Invalid option: -%c\n", argv[arg][1]);
	    exit(0);
	    }
	 }
      }
   if (n_interfaces == 0) {
      en_table[0].type= EN_MT;  n_interfaces =  1;
      }
   if (pkt_init())  /* Initialise just the packet driver(s) */
      exit(1);  /* Problems in starting one of the packet driver(s) */

   receive();
   }

void write_data(void)
{
   FILE *f;
   int j;
   f = fopen("trace.dat", "w");
   for (j = 0; j != n_p; ++j)
      fprintf(f, "%lu %u\n",  p_table[j].us, p_table[j].sz);
   fclose(f);
   }
void receive()
{
   unsigned short k, n;
   unsigned char ch;
   unsigned long tot;

   tot = set_timeout(1);
   for (;;) {
      if (chk_timeout(tot)) {
         tot = set_timeout(1);  /* 1 second */
         printf("%d packets\n", n_p);
         if (n_p == TRACE_SZ) {
            printf("Time to stop !!!\n");
            write_data();
            pkt_release();
            exit(0);
            }
         if (kbhit()) {
            if ((ch = getch()) == 27) {  /* ESC */
               write_data();
               printf("\nShutting down\n");
               // shut down access to packet driver
               //
               pkt_release();
               exit(0);
               }
            }
         }
      }

   }

void au_send_mon(int length, unsigned char far *buffer)
{  return;  }  /* Default version, never used */
