#include <stdio.h>              // printf()
#include <stdlib.h>             // exit()
#include "pci.h"





#define MAX_PCI_DEVICES     16

#define VENDOR_STRING       "FORE"
#define DEVICE_STRING       "PCA200E"
#define DEVICE_DESC         "ATM NIC"
#define VENDOR_ID           FORE_VENDOR_ID
#define DEVICE_ID           FORE_PCA200E_DEVICE_ID




struct PciDevice
{
    unsigned short handle;
    unsigned long base_address;
    unsigned char irq_level;
};




struct PciDevice devices[ MAX_PCI_DEVICES ];


void main( void )
{
    unsigned count, total;
    unsigned short temp16;

    // check whether PCI BIOS is present
    //
    if( !pci_is_bios_present() )
    {
        printf( "Sorry, PCI BIOS is not present.\n" );
        exit( 1 );
    }

    printf( "searching for %s %s %s\n"
        "using vendor ID 0x%4.4X and device ID 0x%4.4X\n\n",
        VENDOR_STRING, DEVICE_STRING, DEVICE_DESC,
        VENDOR_ID, DEVICE_ID
    );

    // first find all the devices with the desired
    // vendor ID and device ID
    //
    for( count=0; count<MAX_PCI_DEVICES; count++ )
    {
        temp16 = pci_get_device_handle( DEVICE_ID,
            VENDOR_ID, count );
        if( RETCODE_NO_MORE_PCI_DEVICES == temp16 )
        {
            printf( "no more PCI devices match those ID's\n" );
            break;
        }

        printf( "handle 0x%4.4X found\n", temp16 );
        devices[ count ].handle = temp16;
    }

    total = count;

    if( count >= MAX_PCI_DEVICES )
        printf( "warning: program maximum of %d devices found;"
            " there may be more on the bus\n", MAX_PCI_DEVICES );
    else
        printf( "A total of %d %s %s %s were found.\n\n", total, VENDOR_STRING, DEVICE_STRING, DEVICE_DESC );

    // now get the memory base address and IRQ level of each one
    //
    for( count=0; count<total; count++ )
    {
        devices[ count ].base_address = pci_get_base_address( devices[count].handle );
        devices[ count ].irq_level = pci_get_irq_level( devices[count].handle );

        printf( "device with handle 0x%4.4X has memory address 0x%8.8lX"
            " and IRQ level %d\n", devices[count].handle,
            devices[count].base_address, devices[count].irq_level
        );
    }

    printf( "\nPCI configuration test done\n" );
}
