/* 1010, Mon 9 Sep 96

   SNMPAUTH.C:  Routines to parse and build community-based
                   authentication headers for SNMP.
            NeTraMet version, based on CMU SNMPv2 (see below).

   Copyright (C) 1992-1998 by Nevil Brownlee,
   ITSS Technology Development, The University of Auckland */

/* $Log: snmpauth.c,v $
 * Revision 1.2  1998/04/30 04:57:13  rtfm
 * Bug fixes, change copyright date
 *
 *
 * Revision 1.1  1998/04/30 04:07:43  rtfm
 * Initial revision
 */

/**********************************************************************
    Copyright 1988, 1989, 1991, 1992 by Carnegie Mellon University
  
			 All Rights Reserved
  
Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  
  
CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#include <stdio.h>  /* Used by ERROR() define */

#if defined(unix)
#include <sys/types.h>
#include <sys/time.h>
#include <netinet/in.h>
#endif

#ifdef vms
#include <in.h>
#endif

#include "ausnmp.h"

#include "asn1.h"
#include "snmp.h"
#include "snmpimpl.h"


u_char *snmp_auth_parse(u_char *data, int *length,
   u_char *sid, int *slen, int *version)
{
   u_char type;
   long dummy;
   
   data = asn_parse_header(data, length, &type);
   if (data == NULL){
      ERROR("bad header");
      return NULL;
      }
   if (type != (ASN_SEQUENCE | ASN_CONSTRUCTOR)){
       ERROR("wrong auth header type");
       return NULL;
      }
   data = asn_parse_int(data, length, &type, &dummy, sizeof dummy);
   if (data == NULL){
      ERROR("bad parse of version");
      return NULL;
      }
   *version = (int)dummy;
   data = asn_parse_string(data, length, &type, sid, slen);
   if (data == NULL){
      ERROR("bad parse of community");
      return NULL;
      }
   sid[*slen] = '\0';
   return (u_char *)data;
   }

u_char *snmp_auth_build(
   u_char  *data,
   int     *length,
   u_char  *sid,
   int     *slen,
   int    *version,
   int     messagelen)
{
   data = asn_build_sequence(data, length,
      (u_char)(ASN_SEQUENCE | ASN_CONSTRUCTOR), messagelen + *slen + 5);
   if (data == NULL){
      ERROR("buildheader");
      return NULL;
      }
   data = asn_build_int(data, length,
      (u_char)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
      (long *)version, sizeof *version);
   if (data == NULL){
      ERROR("buildint");
      return NULL;
      }
   data = asn_build_string(data, length,
      (u_char)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_OCTET_STR),
      sid, *slen);
   if (data == NULL){
      ERROR("buildstring");
      return NULL;
      }
   return (u_char *)data;
   }
